﻿/**
Copyright(c)2008 nutsu.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/**
* 2次ベジェ・3次ベジェの変換
* @author nutsu
* @version 0.33
*/

package com.nutsu.shape {
	import flash.geom.Point;

	public class BezierConvert {
		
		public function BezierConvert() {
			
		}
		
		/**
		 * QBezier to CBezier
		 * @param	quadratic bezier
		 * @return	cubic bezier
		 */
		public static function Q2C( bezier:QBezier ):CBezier{
			var x0:Number = bezier.x0;
			var y0:Number = bezier.y0;
			var x1:Number = bezier.x1;
			var y1:Number = bezier.y1;
			var cx:Number = bezier.cx;
			var cy:Number = bezier.cy;
			return new CBezier( x0, y0, x1, y1,
								(x0 + 2*cx)/3, (y0 + 2*cy)/3,
								(x1 + 2*cx)/3, (y1 + 2*cy)/3 );
		}
		
		/**
		 * CBezier to QBezier
		 * @param	cubic bezier
		 * @return	QBezier[]
		 */
		public static function C2Q( bezier:CBezier, err:Number=0.3 ):Array{
			err = err*err;
			var len:Number = bezier.length;
			var diff:Point = bezier.diff(0.5);
			
			var res0:Array;
			var res1:Array;
			var res2:Array;
			var bs:Array;
			var bs1:Array;
			var bs2:Array;
			
			//同じか評価
			var qb:QBezier = BezierConvert.pC2Q( bezier );
			if( qb!= null && compareDiff(diff, qb.diff(0.5)) && Math.abs(len - qb.length)<err ){
				return [qb];
			}
				
			//2分割評価
			res0 = fC2Q(bezier);
			if( compareDiff(diff, QBezier(res0[0]).diff(1.0)) && 
				Math.abs(len - QBezier(res0[0]).length - QBezier(res0[1]).length)<err ){
				return res0;
			}
			
			//4分割評価
			bs = bezier.split(0.5);
			res0 = fC2Q(bs[0])
			res1 = fC2Q(bs[1]);
			if( compareDiff(CBezier(bs[0]).diff(0.5), QBezier(res0[0]).diff(1.0)) && 
				compareDiff(CBezier(bs[1]).diff(0.5), QBezier(res1[0]).diff(1.0)) && 
				Math.abs(len - QBezier(res0[0]).length - QBezier(res0[1]).length - QBezier(res1[0]).length - QBezier(res1[1]).length )<err ){
				return res0.concat(res1);
			}
			
			//6分割評価
			bs1 = bezier.split(0.333);
			bs2 = CBezier(bs1[1]).split(0.5);
			res0 = fC2Q(bs1[0]);
			res1 = fC2Q(bs2[0]);
			res2 = fC2Q(bs2[1]);
			if( compareDiff(CBezier(bs1[0]).diff(0.5), QBezier(res0[0]).diff(1.0)) && 
				compareDiff(CBezier(bs2[0]).diff(0.5), QBezier(res1[0]).diff(1.0)) && 
				compareDiff(CBezier(bs2[1]).diff(0.5), QBezier(res2[0]).diff(1.0)) && 
				Math.abs(len - QBezier(res0[0]).length - QBezier(res0[1]).length - QBezier(res1[0]).length - QBezier(res1[1]).length - QBezier(res2[0]).length - QBezier(res2[1]).length)<err ){
				return res0.concat(res1).concat(res2);
			}
			
			//8分割（これを最後にしとく）
			bs1 = CBezier(bs[0]).split(0.5);
			bs2 = CBezier(bs[1]).split(0.5);
			res0 = fC2Q(bs1[0]).concat( fC2Q(bs1[1]) ).concat( fC2Q(bs2[0]) ).concat( fC2Q(bs2[1]) );
			return res0;
		}
		
		private static function compareDiff( dc:Point, dp:Point ):Boolean{
			dc.normalize(1.0);
			dp.normalize(1.0);
			return ( Math.abs(dp.x-dc.x)<0.01 && Math.abs(dp.y-dc.y)<0.01 );
		}
		
		/**
		 * 3次ベジェから2つの2次ベジェへ簡易変換
		 * @param	cubic bezier
		 * @return	QBezier[]
		 */
		public static function fC2Q( bezier:CBezier ):Array{
			var x0:Number = bezier.x0;
			var y0:Number = bezier.y0;
			var x1:Number = bezier.x1;
			var y1:Number = bezier.y1;
			var cx0:Number = bezier.cx0;
			var cy0:Number = bezier.cy0;
			var cx1:Number = bezier.cx1;
			var cy1:Number = bezier.cy1;
			//
			var tp:Number = 1-0.25;
			var f0x:Number  = cx0*tp + x0/4;
			var f0y:Number  = cy0*tp + y0/4;
			var f1x:Number  = cx1*tp + x1/4;
			var f1y:Number  = cy1*tp + y1/4;
			var f2x:Number  = f1x*0.5 + f0x*0.5;
			var f2y:Number  = f1y*0.5 + f0y*0.5;
				
			return [ new QBezier(x0,y0,f2x,f2y,f0x,f0y), new QBezier(f2x,f2y,x1,y1,f1x,f1y) ];
		}
		
		/**
		 * 3次ベジェから1つの2次ベジェへ簡易変換
		 * @param	cubic bezier
		 * @return	QBezier
		 */
		private static function pC2Q( bezier:CBezier ):QBezier{
			var x0:Number = bezier.x0;
			var y0:Number = bezier.y0;
			var x1:Number = bezier.x1;
			var y1:Number = bezier.y1;
			var cx0:Number = bezier.cx0;
			var cy0:Number = bezier.cy0;
			var cx1:Number = bezier.cx1;
			var cy1:Number = bezier.cy1;
			//
			var vx0:Number = cx0 - x0;
			var vy0:Number = cy0 - y0;
			var vx1:Number = cx1 - x1;
			var vy1:Number = cy1 - y1;
			var c:Number = vx0*vy1 - vx1*vy0;
			if( c!=0 ){
				var a:Number = ( vx1*y0 - vx1*y1 + vy1*x1 - vy1*x0 )/c;
				var b:Number = ( a*vx0 + x0 - x1 )/vx1;
				if( a>0 && b>0 ){
					return new QBezier( x0, y0, x1, y1, x0 + a*vx0, y0 + a*vy0 );
				}else{
					return null;
				}
			}else{
				return null;
			}
		}
	}
	
}
