﻿/**
Copyright(c)2008 nutsu.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/**
* Cubic Bezier
* @author nutsu
* @version 0.3
*/

package com.nutsu.shape{

	import flash.display.Graphics;
	import flash.geom.Point;
	
	public class CBezier {
		
		//start point
		private var _x0:Number;
		private var _y0:Number;
		//end point
		private var _x1:Number;
		private var _y1:Number;
		//control point 0
		private var _cx0:Number;
		private var _cy0:Number;
		//control point 0
		private var _cx1:Number;
		private var _cy1:Number;
		//length
		private var _length:Number;
		
		public var length_cal_step:int=100;
		
		/**
		 * Constructor
		*/
		public function CBezier( x0_:Number, y0_:Number, x1_:Number, y1_:Number, cx0_:Number, cy0_:Number, cx1_:Number, cy1_:Number ) {
			_x0=x0_, _y0=y0_, _x1=x1_, _y1=y1_, _cx0=cx0_, _cy0=cy0_, _cx1=cx1_, _cy1=cy1_, _length = Number.NaN;
		}
		
		/**
		 * Cunstruct with Point
		 */
		public static function make( p0:Point, p1:Point, ctrl0:Point, ctrl1:Point ):CBezier{
			return new CBezier( p0.x, p0.y, p1.x, p1.y, ctrl0.x, ctrl0.y, ctrl1.x, ctrl1.y );
		}
		
		/**
		 * Clone
		 */
		public function clone():CBezier{
			var c:CBezier = new CBezier(_x0, _y0, _x1, _y1, _cx0, _cy0, _cx1, _cy1 );
			if( isNaN(_length)==false ){
				c._length  = _length;
			}
			return c;
		}
		
		//------------------------------------------------------------------------------F
		
		/**
		 * Bezier Function
		 * @param	t
		 * @return	Point
		 */
		public function f(t:Number):Point{
			var tp:Number = 1.0 - t;
			return new Point( _x0*tp*tp*tp + 3*_cx0*t*tp*tp + 3*_cx1*t*t*tp + _x1*t*t*t,
							  _y0*tp*tp*tp + 3*_cy0*t*tp*tp + 3*_cy1*t*t*tp + _y1*t*t*t );
		}
		
		/**
		 * Diff of Bezier Function
		 * @param	t
		 * @return	Vector
		 */
		public function diff(t:Number):Point{
			return new Point( 3*(_x1-_x0-3*_cx1+3*_cx0)*t*t + 6*(_x0+_cx1-2*_cx0)*t - 3*_x0+3*_cx0,
							  3*(_y1-_y0-3*_cy1+3*_cy0)*t*t + 6*(_y0+_cy1-2*_cy0)*t - 3*_y0+3*_cy0 );
		}
		
		//------------------------------------------------------------------------------CUT
		
		/**
		 * Split Bezier at t.
		 * @param	t( 0 to 1 )
		 * @return	CBezier[]
		 */
		public function split(t:Number):Array{
			if( t<0 || t>1 ){
				throw new ArgumentError("parameter t:0<t<1");
			}
			if( t==0 ){
				return [ null, clone() ];
			}else if( t==1 ){
				return [ clone(), null ];
			}else{
				var tp:Number = 1-t;
				var mx:Number  = _cx0*tp + _cx1*t;
				var my:Number  = _cy0*tp + _cy1*t;
				var ax0:Number = _x0*tp + _cx0*t;
				var ay0:Number = _y0*tp + _cy0*t;
				var ax1:Number = ax0*tp + mx*t;
				var ay1:Number = ay0*tp + my*t;
				var bx1:Number = _cx1*tp + _x1*t;
				var by1:Number = _cy1*tp + _y1*t;
				var bx0:Number = mx*tp + bx1*t;
				var by0:Number = my*tp + by1*t;
				var px:Number  = ax1*tp + bx0*t;
				var py:Number  = ay1*tp + by0*t;
				return [ new CBezier( _x0, _y0, px, py, ax0, ay0, ax1, ay1 ), new CBezier( px, py, _x1, _y1, bx0, by0, bx1, by1 )];
			}
		}
		
		/**
		 * Clip Bezier between t0 to t1.
		 * @param	t0
		 * @param	t1
		 * @return	CBezier
		 */
		public function clip( t0:Number, t1:Number ):CBezier{
			if( t0==t1 ){
				return null;
			}else if( t0==0 || t0==1 ){
				return (split(t1))[ (t0==1)?1:0 ];
			}else if( t1==0 || t1==1 ){
				return (split(t0))[ (t1==1)?1:0 ];
			}
			
			var tp0:Number   = 1-t0;
			var tp1:Number   = 1-t1;
			
			//fx
			var px0:Number = _x0*tp0*tp0*tp0 + 3*_cx0*t0*tp0*tp0 + 3*_cx1*t0*t0*tp0 + _x1*t0*t0*t0;
			var py0:Number = _y0*tp0*tp0*tp0 + 3*_cy0*t0*tp0*tp0 + 3*_cy1*t0*t0*tp0 + _y1*t0*t0*t0;
			var px1:Number = _x0*tp1*tp1*tp1 + 3*_cx0*t1*tp1*tp1 + 3*_cx1*t1*t1*tp1 + _x1*t1*t1*t1;
			var py1:Number = _y0*tp1*tp1*tp1 + 3*_cy0*t1*tp1*tp1 + 3*_cy1*t1*t1*tp1 + _y1*t1*t1*t1;
			
			var mx0:Number  = _cx0*tp0 + _cx1*t0;
			var my0:Number  = _cy0*tp0 + _cy1*t0;
			var bx:Number   = _cx1*tp0 + _x1*t0;
			var by:Number   = _cy1*tp0 + _y1*t0;
			var cx00:Number = mx0*tp0 + bx*t0;
			var cy00:Number = my0*tp0 + by*t0;
			
			var t2:Number   = (t1-t0)/(1.0-t0);
			var tp2:Number  = 1-t2;
			var mx1:Number  = cx00*tp2 + bx*t2;
			var my1:Number  = cy00*tp2 + by*t2;
			
			var nx0:Number = px0*tp2 + cx00*t2;
			var ny0:Number = py0*tp2 + cy00*t2;
			var nx1:Number = nx0*tp2 + mx1*t2;
			var ny1:Number = ny0*tp2 + my1*t2;
			
			return new CBezier( px0, py0, px1, py1, nx0, ny0, nx1, ny1 );
		}
		
		//------------------------------------------------------------------------------LENGTH
		
		public function get length():Number{
			if( isNaN(_length) ){
				var k:Number = 1.0/length_cal_step;
				var px0:Number = _x0;
				var py0:Number = _y0;
				var px1:Number;
				var py1:Number;
				var dx:Number;
				var dy:Number;
				var t:Number;
				var tp:Number;
				_length = 0.0;
				for(var i:int=1;i<=length_cal_step;i++){
					t = i*k;
					tp = 1-t;
					px1 = _x0*tp*tp*tp + 3*_cx0*t*tp*tp + 3*_cx1*t*t*tp + _x1*t*t*t;
					py1 = _y0*tp*tp*tp + 3*_cy0*t*tp*tp + 3*_cy1*t*t*tp + _y1*t*t*t;
					dx  = px1-px0;
					dy  = py1-py0;
					_length += Math.sqrt(dx*dx+dy*dy);
					px0 = px1;
					py0 = py1;
				}
			}
			return _length;
		}
		
		//------------------------------------------------------------------------------COORDINATES
		
		/**
		* offset coordinates
		* @param	x offset
		* @param	y offset
		*/
		public function offset( x:Number, y:Number ):void{
			_x0  += x;
			_y0  += y;
			_x1  += x;
			_y1  += y;
			_cx0 += x;
			_cx0 += y;
			_cx1 += x;
			_cx1 += y;
		}
		
		public function get point0():Point{
			return new Point(_x0,_y0);
		}
		public function get point1():Point{
			return new Point(_x1,_y1);
		}
		public function get control0():Point{
			return new Point(_cx0,_cy0);
		}
		public function get control1():Point{
			return new Point(_cx1,_cy1);
		}
		public function set point0(p:Point):void{
			_x0 = p.x; _y0 = p.y; _length = Number.NaN;
		}
		public function set point1(p:Point):void{
			_x1 = p.x; _y1 = p.y; _length = Number.NaN;
		}
		public function set control0(p:Point):void{
			_cx0 = p.x; _cx0 = p.y; _length = Number.NaN;
		}
		public function set control1(p:Point):void{
			_cx1 = p.x; _cx1 = p.y; _length = Number.NaN;
		}
		//------
		
		public function get x0():Number{
			return _x0;
		}
		public function get y0():Number{
			return _y0;
		}
		public function get x1():Number{
			return _x1;
		}
		public function get y1():Number{
			return _y1;
		}
		public function get cx0():Number{
			return _cx0;
		}
		public function get cy0():Number{
			return _cy0;
		}
		public function get cx1():Number{
			return _cx1;
		}
		public function get cy1():Number{
			return _cy1;
		}
		public function set x0(v:Number):void{
			_x0 = v; _length = Number.NaN;
		}
		public function set y0(v:Number):void{
			_y0 = v; _length = Number.NaN;
		}
		public function set x1(v:Number):void{
			_x1 = v; _length = Number.NaN;
		}
		public function set y1(v:Number):void{
			_y1 = v; _length = Number.NaN;
		}
		public function set cx0(v:Number):void{
			_cx0 = v; _length = Number.NaN;
		}
		public function set cy0(v:Number):void{
			_cy0 = v; _length = Number.NaN;
		}
		public function set cx1(v:Number):void{
			_cx1 = v; _length = Number.NaN;
		}
		public function set cy1(v:Number):void{
			_cy1 = v; _length = Number.NaN;
		}
		//------------------------------------------------------------------------------GRAPHICS
		
		/**
		 * draw
		 * @param	Graphics
		 * @param	draw line number
		 */
		public function draw( g:Graphics, step:int=100 ):void{
			moveTo(g)
			curveTo(g,step);
		}
		
		/**
		 * moveTo
		 * @param	Graphics
		 */
		public function moveTo( g:Graphics ):void{
			g.moveTo( _x0, _y0 );
		}
		
		/**
		 * curveTo
		 * @param	Graphics
		 * @param	draw line number
		 */
		public function curveTo( g:Graphics, step:int=100 ):void{
			var k:Number = 1.0/step;
			var t:Number; var tp:Number; var px:Number; var py:Number;
			for( var i:int=1;i<=step;i++ ){
				t = i*k;
				tp = 1-t;
				px = _x0*tp*tp*tp + 3*_cx0*t*tp*tp + 3*_cx1*t*t*tp + _x1*t*t*t;
				py = _y0*tp*tp*tp + 3*_cy0*t*tp*tp + 3*_cy1*t*t*tp + _y1*t*t*t;
				g.lineTo( px, py );
			}
		}
		
	}
	
}
